#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Validators for v4 datastore protocol buffers.

This module is internal and should not be used by client applications.
"""

















import re

from google.appengine.datastore import datastore_pbs
from google.appengine.datastore import datastore_v4_pb2



_DATASET_ID_PARTITION_STRING = (r'[a-z\d\-]{1,%d}'
                                % datastore_pbs.MAX_DATASET_ID_SECTION_LENGTH)

_DATASET_ID_DOMAIN_STRING = (
    r'[a-z\d][a-z\d\-\.]{0,%d}'
    % (datastore_pbs.MAX_DATASET_ID_SECTION_LENGTH - 1))

_DATASET_ID_DISPLAY_STRING = (
    r'[a-z\d][a-z\d\-]{0,%d}'
    % (datastore_pbs.MAX_DATASET_ID_SECTION_LENGTH - 1))

_DATASET_ID_STRING = ('(?:(?:%s)~)?(?:(?:%s):)?(?:%s)'
                      % (_DATASET_ID_PARTITION_STRING,
                         _DATASET_ID_DOMAIN_STRING,
                         _DATASET_ID_DISPLAY_STRING))


_DATASET_ID_RE = re.compile('^%s$' % _DATASET_ID_STRING)


_RESERVED_NAME_RE = re.compile('^__(.*)__$')


_PARTITION_ID_RE = re.compile(r'^[0-9A-Za-z\._\-]{0,%d}$'
                              % datastore_pbs.MAX_PARTITION_ID_LENGTH)


_STRING_VALUE_MEANINGS = frozenset([
    datastore_pbs.MEANING_TEXT,
    datastore_pbs.MEANING_ATOM_CATEGORY,
    datastore_pbs.MEANING_URL,
    datastore_pbs.MEANING_ATOM_TITLE,
    datastore_pbs.MEANING_ATOM_CONTENT,
    datastore_pbs.MEANING_ATOM_SUMMARY,
    datastore_pbs.MEANING_ATOM_AUTHOR,
    datastore_pbs.MEANING_GD_EMAIL,
    datastore_pbs.MEANING_GD_IM,
    datastore_pbs.MEANING_GD_PHONENUMBER,
    datastore_pbs.MEANING_GD_POSTALADDRESS,
    ])


_BLOB_VALUE_MEANINGS = frozenset([
    datastore_pbs.MEANING_BYTESTRING,
    datastore_pbs.MEANING_ZLIB,
    ])


_ENTITY_VALUE_MEANINGS = frozenset([
    datastore_pbs.MEANING_GEORSS_POINT,
    datastore_pbs.MEANING_PREDEFINED_ENTITY_POINT,
    datastore_pbs.MEANING_PREDEFINED_ENTITY_USER,
    ])


_USER_ENTITY_PROPERTY_MAP = {
    datastore_pbs.PROPERTY_NAME_EMAIL: 'string',
    datastore_pbs.PROPERTY_NAME_AUTH_DOMAIN: 'string',
    datastore_pbs.PROPERTY_NAME_USER_ID: 'string',
    datastore_pbs.PROPERTY_NAME_INTERNAL_ID: 'integer',
    datastore_pbs.PROPERTY_NAME_FEDERATED_IDENTITY: 'string',
    datastore_pbs.PROPERTY_NAME_FEDERATED_PROVIDER: 'string',
    }


_USER_ENTITY_REQUIRED_PROPERTIES = frozenset([
    datastore_pbs.PROPERTY_NAME_EMAIL,
    datastore_pbs.PROPERTY_NAME_AUTH_DOMAIN,
    ])


_POINT_ENTITY_PROPERTY_MAP = {
    datastore_pbs.PROPERTY_NAME_X: 'double',
    datastore_pbs.PROPERTY_NAME_Y: 'double',
    }


_POINT_ENTITY_REQUIRED_PROPERTIES = frozenset([
    datastore_pbs.PROPERTY_NAME_X,
    datastore_pbs.PROPERTY_NAME_Y,
    ])



class ValidationError(Exception):
  """Raised when validation fails."""
  pass


def _assert_condition(condition, message):
  """Asserts a validation condition and raises an error if it's not met.

  Args:
    condition: (boolean) condition to enforce
    message: error message

  Raises:
    ValidationError: if condition is not met
  """
  if not condition:
    raise ValidationError(message)


def _assert_initialized(pb):
  """Asserts that a protocol buffer is initialized.

  Args:
    pb: a protocol buffer

  Raises:
    ValidationError: if protocol buffer is not initialized
  """
  errors = []
  if not pb.IsInitialized(errors):
    _assert_condition(False, '\n\t'.join(errors))


def _assert_valid_utf8(string, desc):
  """Asserts that a string is valid UTF8.

  Args:
    string: string to check
    desc: description of the string (used in error message)

  Raises:
    ValidationError: if the string is not valid UTF8
  """
  _assert_condition(datastore_pbs.is_valid_utf8(string),
                    'The %s is not valid UTF-8.' % desc)


def _assert_string_not_empty(string, desc):
  """Asserts that a string is not empty.

  Args:
    string: string to check
    desc: description of the string (used in error message)

  Raises:
    ValidationError: if the string is empty
  """
  _assert_condition(string, 'The %s is the empty string.' % desc)


def _assert_string_not_reserved(string, desc):
  """Asserts that a string is not a reserved name.

  Args:
    string: string to check
    desc: description of the string (used in error message)

  Raises:
    ValidationError: if the string is a reserved name
  """
  _assert_condition(not _RESERVED_NAME_RE.match(string),
                    'The %s "%s" is reserved.'  % (desc, string))


def _assert_string_not_too_long(string, max_length, desc):
  """Asserts that a string is within the maximum string size bounds.

  Args:
    string: string to check
    max_length: max length of the string (inclusive)
    desc: description of the string (used in error message)

  Raises:
    ValidationError: if the string is a reserved name
  """
  _assert_condition(len(string) <= max_length,
                    'The %s is longer than %d characters.' % (desc, max_length))


class _ValidationConstraint(object):
  """Container for a set of validation constraints."""

  def __init__(self, absent_key_allowed=False, incomplete_key_path_allowed=True,
               complete_key_path_allowed=False, reserved_key_allowed=False,
               reserved_property_name_allowed=False,
               meaning_index_only_allowed=False):
    self.__absent_key_allowed = absent_key_allowed
    self.__incomplete_key_path_allowed = incomplete_key_path_allowed
    self.__complete_key_path_allowed = complete_key_path_allowed
    self.__reserved_key_allowed = reserved_key_allowed
    self.__reserved_property_name_allowed = reserved_property_name_allowed
    self.__meaning_index_only_allowed = meaning_index_only_allowed

  @property
  def absent_key_allowed(self):
    """Allow keys to be absent from entities."""
    return self.__absent_key_allowed

  @property
  def incomplete_key_path_allowed(self):
    """Allow key paths to be incomplete."""
    return self.__incomplete_key_path_allowed

  @property
  def complete_key_path_allowed(self):
    """Allow key paths to be complete."""
    return self.__complete_key_path_allowed

  @property
  def reserved_key_allowed(self):
    """Allow reserved keys and reserved partition ids."""
    return self.__reserved_key_allowed

  @property
  def reserved_property_name_allowed(self):
    """Allow reserved property names."""
    return self.__reserved_property_name_allowed

  @property
  def meaning_index_only_allowed(self):
    """Allow the index only meaning."""
    return self.__meaning_index_only_allowed

  def __hash__(self):
    return hash(id(self))

  def __eq__(self, other):
    return self is other



READ = _ValidationConstraint(
    absent_key_allowed=False,
    incomplete_key_path_allowed=False,
    complete_key_path_allowed=True,
    reserved_key_allowed=True,
    reserved_property_name_allowed=True,
    meaning_index_only_allowed=True)



READ_ENTITY_IN_VALUE = _ValidationConstraint(
    absent_key_allowed=True,
    incomplete_key_path_allowed=True,
    complete_key_path_allowed=True,
    reserved_key_allowed=True,
    reserved_property_name_allowed=True,
    meaning_index_only_allowed=True)


WRITE = _ValidationConstraint(
    absent_key_allowed=False,
    incomplete_key_path_allowed=False,
    complete_key_path_allowed=True,
    reserved_key_allowed=False,
    reserved_property_name_allowed=False,
    meaning_index_only_allowed=False)


WRITE_ENTITY_IN_VALUE = _ValidationConstraint(
    absent_key_allowed=True,
    incomplete_key_path_allowed=True,
    complete_key_path_allowed=True,
    reserved_key_allowed=True,
    reserved_property_name_allowed=False,
    meaning_index_only_allowed=False)


ALLOCATE_KEY_ID = _ValidationConstraint(
    absent_key_allowed=False,
    incomplete_key_path_allowed=True,
    complete_key_path_allowed=False,
    reserved_key_allowed=False,
    reserved_property_name_allowed=False,
    meaning_index_only_allowed=False)


WRITE_AUTO_ID = _ValidationConstraint(
    absent_key_allowed=False,
    incomplete_key_path_allowed=True,
    complete_key_path_allowed=False,
    reserved_key_allowed=False,
    reserved_property_name_allowed=False,
    meaning_index_only_allowed=False)


KEY_IN_VALUE = _ValidationConstraint(
    absent_key_allowed=False,
    incomplete_key_path_allowed=False,
    complete_key_path_allowed=True,
    reserved_key_allowed=True,
    reserved_property_name_allowed=True,
    meaning_index_only_allowed=False)


_ENTITY_IN_VALUE_CONSTRAINTS = {
    READ: READ_ENTITY_IN_VALUE,
    READ_ENTITY_IN_VALUE: READ_ENTITY_IN_VALUE,
    WRITE: WRITE_ENTITY_IN_VALUE,
    WRITE_AUTO_ID: WRITE_ENTITY_IN_VALUE,
    WRITE_ENTITY_IN_VALUE: WRITE_ENTITY_IN_VALUE,
}


def _get_entity_in_value_constraint(constraint):
  """Returns the corresponding constraint for entities in values.

  Args:
    constraint: a _ValidationConstraint

  Returns:
    a _ValidationConstraint for entities in values

  Raises:
    ValueError: if no corresponding constraint exists
  """
  if constraint not in _ENTITY_IN_VALUE_CONSTRAINTS:
    raise ValueError('No corresponding constraint for entities in values.')
  return _ENTITY_IN_VALUE_CONSTRAINTS[constraint]











class _EntityValidator(object):
  """Validator for v4 entities and keys."""

  def validate_keys(self, constraint, keys):
    """Validates a list of keys.

    Args:
      constraint: a _ValidationConstraint to apply
      keys: a list of entity_v4_pb.Key objects

    Raises:
      ValidationError: if any of the keys is invalid
    """
    for key in keys:
      self.validate_key(constraint, key)

  def validate_key(self, constraint, key):
    """Validates a key.

    Args:
      constraint: a _ValidationConstraint to apply
      key: an entity_v4_pb.Key

    Raises:
      ValidationError: if the key is invalid
    """
    _assert_condition(
        key.HasField('partition_id'), 'Key is missing partition id.')
    self.validate_partition_id(constraint, key.partition_id)
    num_key_path_elements = len(key.path_element)
    _assert_condition(num_key_path_elements, 'Key path is empty.')
    _assert_condition((num_key_path_elements
                       <= datastore_pbs.MAX_KEY_PATH_LENGTH),
                      ('Key path has more than %d elements.'
                       % datastore_pbs.MAX_KEY_PATH_LENGTH))
    num_incomplete_elements = 0
    for path_element in key.path_element:
      _assert_valid_utf8(path_element.kind, 'key path kind')
      kind = path_element.kind
      self.validate_kind(constraint, kind)
      has_name = path_element.HasField('name')
      if path_element.HasField('id'):
        _assert_condition(not has_name,
                          'Key path element has both id (%d) and name ("%s").'
                          % (path_element.id, path_element.name))
        _assert_condition(path_element.id,
                          'Key path element has an id of 0.')
      else:
        if has_name:
          _assert_valid_utf8(path_element.name, 'key path name')
          name = path_element.name
          _assert_string_not_empty(name, 'key path name')
          _assert_string_not_too_long(name,
                                      datastore_pbs.MAX_INDEXED_STRING_CHARS,
                                      'key path name')
          if not constraint.reserved_key_allowed:
            _assert_string_not_reserved(name, 'key path name')
        else:
          num_incomplete_elements += 1
    final_element = key.path_element[num_key_path_elements - 1]
    final_element_complete = (final_element.HasField('id') or
                              final_element.HasField('name'))
    if not constraint.complete_key_path_allowed:
      _assert_condition(not final_element_complete,
                        'Key path is complete: %s.'
                        % datastore_pbs.v4_key_to_string(key))
    if not constraint.incomplete_key_path_allowed:
      _assert_condition(final_element_complete,
                        'Key path is incomplete: %s.'
                        % datastore_pbs.v4_key_to_string(key))
    if final_element_complete:
      num_expected_incomplete = 0
    else:
      num_expected_incomplete = 1
    if num_incomplete_elements != num_expected_incomplete:

      _assert_condition(False, 'Key path element is incomplete: %s.'
                        % datastore_pbs.v4_key_to_string(key))

  def validate_partition_id(self, constraint, partition_id):
    """Validates a partition ID.

    Args:
      constraint: a _ValidationConstraint to apply
      partition_id: a datastore_v4_pb2.PartitionId

    Raises:
      ValidationError: if the partition ID is invalid
    """
    _assert_condition(partition_id.HasField('dataset_id'),
                      'Partition id is missing dataset id.')
    if partition_id.HasField('dataset_id'):
      self.validate_dataset_id(constraint, partition_id.dataset_id)
    if partition_id.HasField('namespace'):
      self.validate_partition_id_dimension(constraint, partition_id.namespace,
                                           'namespace')

  def validate_dataset_id(self, constraint, dataset_id):
    """Validates a dataset ID.

    Args:
      constraint: a _ValidationConstraint to apply
      dataset_id: dataset ID

    Raises:
      ValidationError: if the partition ID dimension is invalid
    """
    _assert_valid_utf8(dataset_id, 'dataset id')
    _assert_string_not_empty(dataset_id, 'dataset id')
    _assert_string_not_too_long(dataset_id, datastore_pbs.MAX_DATASET_ID_LENGTH,
                                'dataset id')
    _assert_condition(_DATASET_ID_RE.match(dataset_id),
                      'Illegal string "%s" in dataset id.' % dataset_id)



    if not constraint.reserved_key_allowed:
      _assert_string_not_reserved(dataset_id, 'dataset id')

  def validate_partition_id_dimension(self, constraint, partition_dimension,
                                      desc):
    """Validates a dimension (e.g. namespace) of a partition ID.

    Should not be used for datasets (see validate_dataset).

    Args:
      constraint: a _ValidationConstraint to apply
      partition_dimension: string representing one dimension of a partition ID
      desc: description of the dimension (used in error messages)

    Raises:
      ValidationError: if the partition ID dimension is invalid
    """
    _assert_valid_utf8(partition_dimension, desc)
    _assert_string_not_empty(partition_dimension, desc)
    _assert_string_not_too_long(partition_dimension,
                                datastore_pbs.MAX_PARTITION_ID_LENGTH, desc)
    if not constraint.reserved_key_allowed:
      _assert_string_not_reserved(partition_dimension, desc)
    _assert_condition(_PARTITION_ID_RE.match(partition_dimension),
                      'Illegal string "%s" in %s.' % (partition_dimension,
                                                      desc))

  def validate_kind(self, constraint, kind):
    """Validates a kind.

    Args:
      constraint: a _ValidationConstraint to apply
      kind: kind string

    Raises:
      ValidationError: if the kind is invalid
    """
    _assert_string_not_empty(kind, 'kind')
    _assert_string_not_too_long(kind, datastore_pbs.MAX_INDEXED_STRING_CHARS,
                                'kind')
    if not constraint.reserved_key_allowed:
      _assert_string_not_reserved(kind, 'kind')

  def validate_entities(self, constraint, entities):
    """Validates a list of entities.

    Args:
      constraint: a _ValidationConstraint to apply
      entities: a list of entity_v4_pb.Entity objects

    Raises:
      ValidationError: if any of the entities is invalid
    """
    for entity in entities:
      self.validate_entity(constraint, entity)

  def validate_entity(self, constraint, entity):
    """Validates an entity.

    Args:
      constraint: a _ValidationConstraint to apply
      entity: an entity_v4_pb.Entity

    Raises:
      ValidationError: if the entity is invalid
    """
    if entity.HasField('key'):
      self.validate_key(constraint, entity.key)
    else:
      _assert_condition(constraint.absent_key_allowed,
                        'Entity is missing key.')
    property_names = set()
    for prop in entity.property:
      property_name = prop.name
      _assert_condition(property_name not in property_names,
                        ('Entity has duplicate property name "%s".'
                         % property_name))
      property_names.add(property_name)
      self.validate_property(constraint, prop)

  def validate_property(self, constraint, prop):
    """Validates a property.

    Args:
      constraint: a _ValidationConstraint to apply
      prop: an entity_v4_pb.Property

    Raises:
      ValidationError: if the property is invalid
    """
    _assert_valid_utf8(prop.name, 'property name')
    property_name = prop.name
    self.validate_property_name(constraint, property_name)

    _assert_condition(not prop.HasField('deprecated_multi'),
                      'deprecated_multi field not supported.')
    _assert_condition(not prop.deprecated_value,
                      'deprecated_value field not supported.')
    _assert_condition(prop.HasField('value'),
                      'Property "%s" has no value.' % property_name)
    self.validate_value(constraint, prop.value)

  def validate_value(self, constraint, value):
    """Validates a value.

    Args:
      constraint: a _ValidationConstraint to apply
      value: an entity_v4_pb.Value

    Raises:
      ValidationError: if the value is invalid
    """
    self.__validate_value_union(value)
    if value.HasField('string_value'):
      _assert_valid_utf8(value.string_value, 'string value')
    elif value.HasField('blob_key_value'):
      _assert_valid_utf8(value.blob_key_value, 'blob key value')
    elif value.HasField('key_value'):
      self.validate_key(KEY_IN_VALUE, value.key_value)
    elif value.HasField('entity_value'):
      entity_in_value_constraint = _get_entity_in_value_constraint(constraint)
      self.validate_entity(entity_in_value_constraint, value.entity_value)
    elif value.list_value:
      _assert_condition(not value.HasField('indexed'),
                        ('A Value containing a list_value cannot specify '
                         'indexed.'))
      _assert_condition(not value.HasField('meaning'),
                        ('A Value containing a list_value cannot specify '
                         'a meaning.'))
      for sub_value in value.list_value:
        _assert_condition(not sub_value.list_value,
                          ('list_value cannot contain a Value containing '
                           'another list_value.'))
        self.validate_value(constraint, sub_value)
    self.__validate_value_meaning_matches_union(value)
    self.__validate_value_meaning_constraints(constraint, value)
    self.__validate_value_index_constraints(value)

  def __validate_value_union(self, value):
    """Validates that a value is a valid union.

    Args:
      value: an entity_v4_pb.Value

    Raises:
      ValidationError: if the value contains more than one type
    """
    num_sub_values = (value.HasField('boolean_value')
                      + value.HasField('integer_value')
                      + value.HasField('double_value')
                      + value.HasField('timestamp_microseconds_value')
                      + value.HasField('key_value')
                      + value.HasField('blob_key_value')
                      + value.HasField('string_value')
                      + value.HasField('blob_value')
                      + value.HasField('entity_value')
                      + bool(value.list_value))
    _assert_condition(num_sub_values <= 1,
                      'Value has multiple <type>_value fields set.')
    return num_sub_values

  def __validate_value_meaning_matches_union(self, value):
    """Validates that a value's meaning matches its value type.

    Args:
      value: an entity_v4_pb.Value

    Raises:
      ValidationError: if the Value's value type does not match its meaning
    """
    if not value.HasField('meaning'):
      return
    message = 'Value meaning %d does not match %s field.'
    meaning = value.meaning
    if meaning in _STRING_VALUE_MEANINGS:
      _assert_condition(value.HasField('string_value'),
                        message % (meaning, 'string_value'))
    elif meaning in _BLOB_VALUE_MEANINGS:
      _assert_condition(value.HasField('blob_value'),
                        message % (meaning, 'blob_value'))
    elif meaning in _ENTITY_VALUE_MEANINGS:
      _assert_condition(value.HasField('entity_value'),
                        message % (meaning, 'entity_value'))
    elif meaning == datastore_pbs.MEANING_PERCENT:
      _assert_condition(value.HasField('integer_value'),
                        message % (meaning, 'integer_value'))
    elif meaning == datastore_pbs.MEANING_INDEX_ONLY:
      _assert_condition(not value.HasField('timestamp_microseconds_value'),
                        message % (meaning, 'timestamp_microseconds_value'))
      _assert_condition(not value.HasField('blob_key_value'),
                        message % (meaning, 'blob_key_value'))
      _assert_condition(not value.HasField('entity_value'),
                        message % (meaning, 'entity_value'))
    elif meaning == datastore_pbs.MEANING_EMPTY_LIST:
      _assert_condition(self.__validate_value_union(value) == 0,
                        'Empty list cannot have any value fields set.')
    else:
      _assert_condition(False,
                        'Unknown value meaning %d' % meaning)

  def __validate_value_meaning_constraints(self, constraint, value):
    """Checks constraints on values that result from their meaning.

    For example, some meanings cause the length of a value to be constrained.

    Args:
      constraint: a _ValidationConstraint to apply
      value: an entity_v4_pb.Value

    Raises:
      ValidationError: if the value is invalid
    """
    if not value.HasField('meaning'):
      return
    meaning = value.meaning
    if meaning == datastore_pbs.MEANING_BYTESTRING:
      _assert_condition((len(value.blob_value)
                         <= datastore_pbs.MAX_INDEXED_BLOB_BYTES),
                        ('Blob value with meaning %d has more than '
                         'permitted %d bytes.'
                         % (meaning, datastore_pbs.MAX_INDEXED_BLOB_BYTES)))
    elif meaning in (datastore_pbs.MEANING_TEXT, datastore_pbs.MEANING_ZLIB):
      _assert_condition(not value.indexed,
                        'Indexed value has meaning %d.' % meaning)
    elif meaning == datastore_pbs.MEANING_URL:
      _assert_condition((len(value.string_value)
                         <= datastore_pbs.MAX_URL_CHARS),
                        'URL value has more than permitted %d characters.'
                        % datastore_pbs.MAX_URL_CHARS)
    elif meaning == datastore_pbs.MEANING_PERCENT:
      _assert_condition((value.integer_value >= 0
                         and value.integer_value <= 100),
                        'Percent value outside permitted range [0, 100].')
    elif meaning == datastore_pbs.MEANING_GEORSS_POINT:
      property_map = self._validate_predefined_entity_value(
          value.entity_value, 'geo point', _POINT_ENTITY_PROPERTY_MAP,
          _POINT_ENTITY_REQUIRED_PROPERTIES)
      latitude = property_map[datastore_pbs.PROPERTY_NAME_X].double_value
      longitude = property_map[datastore_pbs.PROPERTY_NAME_Y].double_value
      _assert_condition(abs(latitude) <= 90.0,
                        'Latitude outside permitted range [-90.0, 90.0].')
      _assert_condition(abs(longitude) <= 180.0,
                        'Longitude outside permitted range [-180.0, 180.0].')
    elif meaning == datastore_pbs.MEANING_PREDEFINED_ENTITY_POINT:
      self._validate_predefined_entity_value(value.entity_value, 'point',
                                             _POINT_ENTITY_PROPERTY_MAP,
                                             _POINT_ENTITY_REQUIRED_PROPERTIES)
    elif meaning == datastore_pbs.MEANING_PREDEFINED_ENTITY_USER:
      self._validate_predefined_entity_value(value.entity_value, 'user',
                                             _USER_ENTITY_PROPERTY_MAP,
                                             _USER_ENTITY_REQUIRED_PROPERTIES)
    elif meaning == datastore_pbs.MEANING_INDEX_ONLY:
      _assert_condition(constraint.meaning_index_only_allowed,
                        'Value has meaning %d.'
                        % datastore_pbs.MEANING_INDEX_ONLY)

  def _validate_predefined_entity_value(self, entity, entity_name,
                                        allowed_property_map,
                                        required_properties):
    """Validates a predefined entity (e.g. a user or a point).

    Args:
      entity: the predefined entity (an entity_v4_pb.Entity)
      entity_name: the name of the entity (used in error messages)
      allowed_property_map: a dict whose keys are property names allowed in
          the entity and values are the expected types of these properties
      required_properties: a list of required property names

    Returns:
      a dict of entity_v4_pb2.Value objects keyed by property name

    Raises:
      ValidationError: if the entity is invalid
    """
    _assert_condition(not entity.HasField('key'),
                      'The %s entity has a key.' % entity_name)
    property_map = {}
    for prop in entity.property:
      property_name = prop.name
      _assert_condition(property_name in allowed_property_map,
                        'The %s entity property "%s" is not allowed.'
                        % (entity_name, property_name))
      value = prop.value
      hasser = '%s_value' % allowed_property_map[property_name]
      _assert_condition(
          value.HasField(hasser),
          ('The %s entity property "%s" is the wrong type.'
           % (entity_name, property_name)))
      _assert_condition(not value.HasField('meaning'),
                        'The %s entity property "%s" has a meaning.'
                        % (entity_name, property_name))
      _assert_condition(not value.indexed,
                        'The %s entity property "%s" is indexed.'
                        % (entity_name, property_name))
      property_map[property_name] = value
    for required_property_name in required_properties:
      _assert_condition(required_property_name in property_map,
                        'The %s entity is missing required property "%s".'
                        % (entity_name, required_property_name))
    return property_map

  def __validate_value_index_constraints(self, value):
    """Checks constraints on values that result from their being indexed.

    Args:
      value: an entity_v4_pb.Value

    Raises:
      ValidationError: if the value is invalid
    """
    if not value.indexed:
      return
    if (value.HasField('string_value')
        and value.meaning != datastore_pbs.MEANING_URL):

      _assert_condition((len(value.string_value)
                         <= datastore_pbs.MAX_INDEXED_STRING_CHARS),
                        ('Indexed string value has more than %d permitted '
                         'characters.'
                         % datastore_pbs.MAX_INDEXED_STRING_CHARS))
    elif value.HasField('blob_value'):
      _assert_condition((len(value.blob_value)
                         <= datastore_pbs.MAX_INDEXED_BLOB_BYTES),
                        ('Indexed blob value has more than %d permitted '
                         'bytes.' % datastore_pbs.MAX_INDEXED_BLOB_BYTES))
    elif value.HasField('entity_value'):
      _assert_condition(value.HasField('meaning'),
                        'Entity value is indexed.')

  def validate_property_name(self, constraint, property_name):
    """Validates a property name.

    Args:
      constraint: a _ValidationConstraint to apply
      property_name: name of a property

    Raises:
      ValidationError: if the property name is invalid
    """
    desc = 'property.name'
    _assert_string_not_empty(property_name, desc)
    _assert_string_not_too_long(property_name,
                                datastore_pbs.MAX_INDEXED_STRING_CHARS, desc)


    if not constraint.reserved_property_name_allowed:
      _assert_string_not_reserved(property_name, desc)



__entity_validator = _EntityValidator()


def get_entity_validator():
  """Validator for entities and keys."""
  return __entity_validator



class _QueryValidator(object):
  """Validator for queries."""

  def __init__(self, entity_validator):
    self.__entity_validator = entity_validator

  def validate_query(self, query, is_strong_read_consistency):
    """Validates a Query.

    Args:
      query: a datastore_v4_pb2.Query
      is_strong_read_consistency: whether the request containing the query
          requested strong read consistency

    Raises:
      ValidationError: if the query is invalid
    """
    _assert_condition((not is_strong_read_consistency
                       or self._has_ancestor(query.filter)),
                      'Global queries do not support strong consistency.')
    if query.HasField('filter'):
      self.validate_filter(query.filter)
    for kind_expression in query.kind:
      self.__validate_kind_expression(kind_expression)
    group_by_properties = set()
    for property_reference in query.group_by:
      self.__validate_property_reference(property_reference)
      group_by_properties.add(property_reference.name)
    for property_expression in query.projection:
      self.__validate_property_expression(property_expression,
                                          group_by_properties)
    for property_order in query.order:
      self.__validate_property_order(property_order)

  def validate_filter(self, filt):
    """Validates a Filter.

    Args:
      filt: a datastore_v4_pb2.Filter

    Raises:
      ValidationError: if the filter is invalid
    """
    _assert_condition((filt.HasField('composite_filter')
                       + filt.HasField('property_filter') == 1),
                      'A filter must have exactly one of its fields set.')
    if filt.HasField('composite_filter'):
      comp_filter = filt.composite_filter
      _assert_condition(comp_filter.filter,
                        'A composite filter must have at least one sub-filter.')
      for sub_filter in comp_filter.filter:
        self.validate_filter(sub_filter)
    elif filt.HasField('property_filter'):
      prop_filter = filt.property_filter
      self.__validate_property_reference(prop_filter.property)
      _assert_condition(prop_filter.value.indexed,
                        'A filter value must be indexed.')
      self.__entity_validator.validate_value(READ,
                                             prop_filter.value)

  def __validate_kind_expression(self, kind_expression):
    """Validates a KindExpression.

    Args:
      kind_expression: a datastore_v4_pb2.KindExpression

    Raises:
      ValidationError: if the kind expression is invalid
    """
    _assert_valid_utf8(kind_expression.name, 'kind')
    self.__entity_validator.validate_kind(READ,
                                          kind_expression.name)

  def __validate_property_reference(self, property_reference):
    """Validates a PropertyReference.

    Args:
      property_reference: a datastore_v4_pb2.PropertyReference

    Raises:
      ValidationError: if the property reference is invalid
    """
    _assert_valid_utf8(property_reference.name, 'property name')
    self.__entity_validator.validate_property_name(READ,
                                                   property_reference.name)

  def __validate_property_expression(self, property_expression,
                                     group_by_properties):
    """Validates a PropertyExpression.

    Args:
      property_expression: a datastore_v4_pb2.PropertyExpression
      group_by_properties: the set of property names specified as group by
          properties for the query

    Raises:
      ValidationError: if the property expression is invalid
    """
    self.__validate_property_reference(property_expression.property)
    if not group_by_properties:
      _assert_condition(
          not property_expression.HasField('aggregation_function'),
          'Aggregation function is not allowed without group by.')
    elif property_expression.property.name in group_by_properties:
      _assert_condition(
          not property_expression.HasField('aggregation_function'),
          ('Aggregation function is not allowed for properties '
           'in group by: %s.' % property_expression.property.name))
    else:
      _assert_condition(property_expression.HasField('aggregation_function'),
                        ('Aggregation function is required for properties '
                         'not in group by: %s.'
                         % property_expression.property.name))

  def __validate_property_order(self, property_order):
    """Validates a PropertyOrder.

    Args:
      property_order: a datastore_v4_pb2.PropertyOrder

    Raises:
      ValidationError: if the property expression is invalid
    """
    self.__validate_property_reference(property_order.property)

  def _has_ancestor(self, filt):
    """Determines if a filter includes an ancestor filter.

    Args:
      filt: a datastore_v4_pb2.Filter

    Returns:
      True if the filter includes an ancestor filter, False otherwise
    """
    if filt.HasField('property_filter'):
      op = filt.property_filter.operator
      name = filt.property_filter.property.name
      return (op == datastore_v4_pb2.PropertyFilter.HAS_ANCESTOR
              and name == datastore_pbs.PROPERTY_NAME_KEY)
    if filt.HasField('composite_filter'):
      if (filt.composite_filter.operator
          == datastore_v4_pb2.CompositeFilter.AND):
        for sub_filter in filt.composite_filter.filter:
          if self._has_ancestor(sub_filter):
            return True
    return False



__query_validator = _QueryValidator(__entity_validator)


def get_query_validator():
  """Validator for queries."""
  return __query_validator


class _ServiceValidator(object):
  """Validator for request/response protos."""

  def __init__(self, entity_validator, query_validator):
    self.__entity_validator = entity_validator
    self.__query_validator = query_validator

  def validate_begin_transaction_req(self, req):
    """Validates a normalized BeginTransactionRequest.

    Args:
      req: a datastore_v4_pb2.BeginTransactionRequest

    Raises:
      ValidationError: if the request is invalid
    """
    _assert_initialized(req)

  def validate_rollback_req(self, req):
    """Validates a normalized RunQueryRequest.

    Args:
      req: a datastore_v4_pb2.RunQueryRequest

    Raises:
      ValidationError: if the request is invalid
    """
    _assert_initialized(req)

  def validate_commit_req(self, req):
    """Validates a normalized CommitRequest.

    Args:
      req: a datastore_v4_pb2.CommitRequest

    Raises:
      ValidationError: if the request is invalid
    """
    _assert_initialized(req)
    if req.mode == datastore_v4_pb2.CommitRequest.TRANSACTIONAL:
      _assert_condition(req.HasField('transaction'),
                        'Transactional commit requires a transaction.')
    elif req.mode == datastore_v4_pb2.CommitRequest.NON_TRANSACTIONAL:
      _assert_condition(not req.HasField('transaction'),
                        ('Non-transactional commit cannot specify a '
                         'transaction.'))
    else:
      _assert_condition(False,
                        'Unknown commit mode: %d.' % req.mode)
    self.__validate_deprecated_mutation(req.deprecated_mutation)

  def validate_run_query_req(self, req):
    """Validates a normalized RunQueryRequest.

    Args:
      req: a normalized datastore_v4_pb2.RunQueryRequest

    Raises:
      ValidationError: if the request is invalid
    """



    _assert_condition(not req.HasField('gql_query'), 'GQL not supported.')
    _assert_initialized(req)
    self.validate_read_options(req.read_options)
    self.__entity_validator.validate_partition_id(READ,
                                                  req.partition_id)
    _assert_condition(req.HasField('query'),
                      ('One of fields Query.query and Query.gql_query '
                       'must be set.'))
    self.__query_validator.validate_query(
        req.query,
        (req.read_options.read_consistency
         == datastore_v4_pb2.ReadOptions.STRONG))

  def validate_continue_query_req(self, req):
    _assert_initialized(req)

  def validate_lookup_req(self, req):
    """Validates a LookupRequest.

    Args:
      req: a datastore_v4_pb2.LookupRequest

    Raises:
      ValidationError: if the request is invalid
    """
    _assert_initialized(req)
    self.validate_read_options(req.read_options)
    self.__entity_validator.validate_keys(READ, req.key)

  def validate_allocate_ids_req(self, req):
    """Validates an AllocateIdsRequest.

    Args:
      req: a datastore_v4_pb2.AllocateIdsRequest

    Raises:
      ValidationError: if the request is invalid
    """
    _assert_initialized(req)
    _assert_condition(not req.allocate or not req.reserve,
                      'Cannot reserve and allocate ids in the same request.')
    self.__entity_validator.validate_keys(ALLOCATE_KEY_ID,
                                          req.allocate)
    self.__entity_validator.validate_keys(WRITE,
                                          req.reserve)

  def validate_read_options(self, read_options):
    _assert_condition((not read_options.HasField('read_consistency')
                       or not read_options.HasField('transaction')),
                      ('Cannot specify both a read consistency and'
                       ' a transaction.'))

  def __validate_deprecated_mutation(self, deprecated_mutation):
    self.__entity_validator.validate_entities(WRITE,
                                              deprecated_mutation.upsert)
    self.__entity_validator.validate_entities(WRITE,
                                              deprecated_mutation.update)
    self.__entity_validator.validate_entities(WRITE,
                                              deprecated_mutation.insert)
    self.__entity_validator.validate_entities(
        WRITE_AUTO_ID,
        deprecated_mutation.insert_auto_id)
    self.__entity_validator.validate_keys(WRITE,
                                          deprecated_mutation.delete)



__service_validator = _ServiceValidator(__entity_validator,
                                        __query_validator)


def get_service_validator():
  """Returns a validator for v4 service request/response protos.

  Returns:
    a _ServiceValidator
  """
  return __service_validator



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Mixin to assert whether two images are equal."""

import io

import google
from PIL import Image
import six
from six.moves import range


class ImageComparisonMixin(object):

  def assert_images_are_equal(self, expected_bytes, actual_bytes):
    """Assert the images are equal pixel by pixel.

    The image comparison is pixel by pixel. In case both pixels are fully
    transparent then we consider them equal regardless of the color values.

    Note: metadata (EXIF, color profile) is not checked in this check.

    Args:
      expected_bytes: bytes representing the expected image
      actual_bytes: bytes representing the actual image
    """
    self.assertTrue(
        _images_are_equal(expected_bytes, actual_bytes),
        'Images differ:\nExpected:\n%r\nActual:\n%r' % (expected_bytes,
                                                        actual_bytes))


def _images_are_equal(bytes1, bytes2):
  """Check whether two images are equal byte by byte.

  The image comparison is pixel by pixel. In case both pixels are fully
  transparent then we consider them equal regardless of the color values.

  Note: metadata (EXIF, color profile) is not checked in this check.

  Args:
    bytes1: bytes representing the first image
    bytes2: bytes representing the second image

  Returns:
    True if the images are equal using above definition, False otherwise.
  """





  image1 = Image.open(io.BytesIO(bytes1)).convert('RGBA')
  image2 = Image.open(io.BytesIO(bytes2)).convert('RGBA')

  img1_bytes = _tobytes(image1)
  img2_bytes = _tobytes(image2)

  if len(img1_bytes) != len(img2_bytes):
    return False




  null_byte = b'\x00' [0]
  for i in range(len(img1_bytes) // 4):
    pos = 4 * i
    if img1_bytes[pos + 3] == null_byte and img2_bytes[pos + 3] == null_byte:
      continue

    if img1_bytes[pos:pos + 4] != img2_bytes[pos:pos + 4]:
      return False

  return True


def _tobytes(image):
  func = image.tostring if six.PY2 else image.tobytes
  return func('raw', 'RGBA')



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#



"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


_sym_db = _symbol_database.Default()


from google.appengine.datastore import action_pb2 as google_dot_appengine_dot_datastore_dot_action__pb2
from google.appengine.datastore import entity_bytes_pb2 as google_dot_appengine_dot_datastore_dot_entity__bytes__pb2
from google.appengine.datastore import snapshot_pb2 as google_dot_appengine_dot_datastore_dot_snapshot__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n3google/appengine/datastore/datastore_v3_bytes.proto\x12\x1d\x61pphosting_datastore_v3_bytes\x1a\'google/appengine/datastore/action.proto\x1a-google/appengine/datastore/entity_bytes.proto\x1a)google/appengine/datastore/snapshot.proto\"\xa0\x01\n\x0bTransaction\x12\x0e\n\x06handle\x18\x01 \x02(\x06\x12\x0b\n\x03\x61pp\x18\x02 \x02(\t\x12\x13\n\x0b\x64\x61tabase_id\x18\x06 \x01(\t\x12\x1b\n\x0cmark_changes\x18\x03 \x01(\x08:\x05\x66\x61lse\x12\x42\n\x0f\x63omposite_index\x18\x05 \x03(\x0b\x32).storage_onestore_v3_bytes.CompositeIndex\"\xe1\x0c\n\x05Query\x12\x0b\n\x03\x61pp\x18\x01 \x02(\t\x12\x13\n\x0b\x64\x61tabase_id\x18* \x01(\t\x12\x12\n\nname_space\x18\x1d \x01(\t\x12\x0c\n\x04kind\x18\x03 \x01(\t\x12\x36\n\x08\x61ncestor\x18\x11 \x01(\x0b\x32$.storage_onestore_v3_bytes.Reference\x12\x0f\n\x07shallow\x18+ \x01(\x08\x12;\n\x06\x66ilter\x18\x04 \x03(\n2+.apphosting_datastore_v3_bytes.Query.Filter\x12\x39\n\x05order\x18\t \x03(\n2*.apphosting_datastore_v3_bytes.Query.Order\x12\r\n\x05\x63ount\x18\x17 \x01(\x05\x12\x11\n\x06offset\x18\x0c \x01(\x05:\x01\x30\x12\r\n\x05limit\x18\x10 \x01(\x05\x12\x46\n\x0f\x63ompiled_cursor\x18\x1e \x01(\x0b\x32-.apphosting_datastore_v3_bytes.CompiledCursor\x12J\n\x13\x65nd_compiled_cursor\x18\x1f \x01(\x0b\x32-.apphosting_datastore_v3_bytes.CompiledCursor\x12\x42\n\x0f\x63omposite_index\x18\x13 \x03(\x0b\x32).storage_onestore_v3_bytes.CompositeIndex\x12\x18\n\tkeys_only\x18\x15 \x01(\x08:\x05\x66\x61lse\x12?\n\x0btransaction\x18\x16 \x01(\x0b\x32*.apphosting_datastore_v3_bytes.Transaction\x12\x13\n\x0b\x66\x61ilover_ms\x18\x1a \x01(\x03\x12\x0e\n\x06strong\x18  \x01(\x08\x12\x15\n\rproperty_name\x18! \x03(\t\x12\x1e\n\x16group_by_property_name\x18\" \x03(\t\x12\x10\n\x08\x64istinct\x18\x18 \x01(\x08\x12\x1d\n\x15min_safe_time_seconds\x18# \x01(\x03\x12\x19\n\x11safe_replica_name\x18$ \x03(\t\x12\x18\n\x0cread_time_us\x18, \x01(\x03\x42\x02\x18\x01\x12\x1e\n\x16read_time_epoch_micros\x18- \x01(\x03\x12;\n\x04hint\x18\x12 \x01(\x0e\x32).apphosting_datastore_v3_bytes.Query.HintB\x02\x18\x01\x12\x18\n\x0csearch_query\x18\x08 \x01(\tB\x02\x18\x01\x12\'\n\x14require_perfect_plan\x18\x14 \x01(\x08:\x05\x66\x61lseB\x02\x18\x01\x12\x1a\n\x07\x63ompile\x18\x19 \x01(\x08:\x05\x66\x61lseB\x02\x18\x01\x12 \n\x0epersist_offset\x18% \x01(\x08:\x04trueB\x02\x18\x01\x1a\xf3\x02\n\x06\x46ilter\x12@\n\x02op\x18\x06 \x02(\x0e\x32\x34.apphosting_datastore_v3_bytes.Query.Filter.Operator\x12\x35\n\x08property\x18\x0e \x03(\x0b\x32#.storage_onestore_v3_bytes.Property\x12<\n\ngeo_region\x18( \x01(\x0b\x32(.apphosting_datastore_v3_bytes.GeoRegion\"\xb1\x01\n\x08Operator\x12\r\n\tLESS_THAN\x10\x01\x12\x16\n\x12LESS_THAN_OR_EQUAL\x10\x02\x12\x10\n\x0cGREATER_THAN\x10\x03\x12\x19\n\x15GREATER_THAN_OR_EQUAL\x10\x04\x12\t\n\x05\x45QUAL\x10\x05\x12\x06\n\x02IN\x10\x06\x12\n\n\x06\x45XISTS\x10\x07\x12\x17\n\x13\x43ONTAINED_IN_REGION\x10\x08\x12\r\n\tNOT_EQUAL\x10\t\x12\n\n\x06NOT_IN\x10\n\x1a\x99\x01\n\x05Order\x12\x10\n\x08property\x18\n \x02(\t\x12R\n\tdirection\x18\x0b \x01(\x0e\x32\x34.apphosting_datastore_v3_bytes.Query.Order.Direction:\tASCENDING\"*\n\tDirection\x12\r\n\tASCENDING\x10\x01\x12\x0e\n\nDESCENDING\x10\x02\"=\n\x04Hint\x12\x0f\n\x0bORDER_FIRST\x10\x01\x12\x12\n\x0e\x41NCESTOR_FIRST\x10\x02\x12\x10\n\x0c\x46ILTER_FIRST\x10\x03\"2\n\x0bRegionPoint\x12\x10\n\x08latitude\x18\x01 \x02(\x01\x12\x11\n\tlongitude\x18\x02 \x02(\x01\"a\n\x0c\x43ircleRegion\x12:\n\x06\x63\x65nter\x18\x01 \x02(\x0b\x32*.apphosting_datastore_v3_bytes.RegionPoint\x12\x15\n\rradius_meters\x18\x02 \x02(\x01\"\x8f\x01\n\x0fRectangleRegion\x12=\n\tsouthwest\x18\x01 \x02(\x0b\x32*.apphosting_datastore_v3_bytes.RegionPoint\x12=\n\tnortheast\x18\x02 \x02(\x0b\x32*.apphosting_datastore_v3_bytes.RegionPoint\"\x8b\x01\n\tGeoRegion\x12;\n\x06\x63ircle\x18\x01 \x01(\x0b\x32+.apphosting_datastore_v3_bytes.CircleRegion\x12\x41\n\trectangle\x18\x02 \x01(\x0b\x32..apphosting_datastore_v3_bytes.RectangleRegion\"\xec\x06\n\rCompiledQuery\x12M\n\x0bprimaryscan\x18\x01 \x02(\n28.apphosting_datastore_v3_bytes.CompiledQuery.PrimaryScan\x12Q\n\rmergejoinscan\x18\x07 \x03(\n2:.apphosting_datastore_v3_bytes.CompiledQuery.MergeJoinScan\x12\x33\n\tindex_def\x18\x15 \x01(\x0b\x32 .storage_onestore_v3_bytes.Index\x12\x11\n\x06offset\x18\n \x01(\x05:\x01\x30\x12\r\n\x05limit\x18\x0b \x01(\x05\x12\x11\n\tkeys_only\x18\x0c \x02(\x08\x12\x15\n\rproperty_name\x18\x18 \x03(\t\x12\x1b\n\x13\x64istinct_infix_size\x18\x19 \x01(\x05\x12\x17\n\x0fkey_path_length\x18\x1b \x01(\x05\x12O\n\x0c\x65ntityfilter\x18\r \x01(\n29.apphosting_datastore_v3_bytes.CompiledQuery.EntityFilter\x12\x12\n\nplan_label\x18\x1a \x01(\t\x1a\xd5\x01\n\x0bPrimaryScan\x12\x12\n\nindex_name\x18\x02 \x01(\t\x12\x11\n\tstart_key\x18\x03 \x01(\x0c\x12\x17\n\x0fstart_inclusive\x18\x04 \x01(\x08\x12\x0f\n\x07\x65nd_key\x18\x05 \x01(\x0c\x12\x15\n\rend_inclusive\x18\x06 \x01(\x08\x12\x1b\n\x13start_postfix_value\x18\x16 \x03(\x0c\x12\x19\n\x11\x65nd_postfix_value\x18\x17 \x03(\x0c\x12&\n\x1e\x65nd_unapplied_log_timestamp_us\x18\x13 \x01(\x03\x1aV\n\rMergeJoinScan\x12\x12\n\nindex_name\x18\x08 \x02(\t\x12\x14\n\x0cprefix_value\x18\t \x03(\x0c\x12\x1b\n\x0cvalue_prefix\x18\x14 \x01(\x08:\x05\x66\x61lse\x1am\n\x0c\x45ntityFilter\x12\x17\n\x08\x64istinct\x18\x0e \x01(\x08:\x05\x66\x61lse\x12\x0c\n\x04kind\x18\x11 \x01(\t\x12\x36\n\x08\x61ncestor\x18\x12 \x01(\x0b\x32$.storage_onestore_v3_bytes.Reference\"\xbe\x04\n\x0e\x43ompiledCursor\x12L\n\x08position\x18\x02 \x01(\n26.apphosting_datastore_v3_bytes.CompiledCursor.PositionB\x02\x18\x01\x12\x41\n\x10postfix_position\x18\x01 \x01(\x0b\x32\'.storage_onestore_v3_bytes.IndexPostfix\x12\x43\n\x11\x61\x62solute_position\x18\x03 \x01(\x0b\x32(.storage_onestore_v3_bytes.IndexPosition\x1a\xd5\x02\n\x08Position\x12\x15\n\tstart_key\x18\x1b \x01(\x0c\x42\x02\x18\x01\x12Y\n\nindexvalue\x18\x1d \x03(\n2A.apphosting_datastore_v3_bytes.CompiledCursor.Position.IndexValueB\x02\x18\x01\x12\x35\n\x03key\x18  \x01(\x0b\x32$.storage_onestore_v3_bytes.ReferenceB\x02\x18\x01\x12!\n\x0fstart_inclusive\x18\x1c \x01(\x08:\x04trueB\x02\x18\x01\x12\x1c\n\x10\x62\x65\x66ore_ascending\x18! \x01(\x08\x42\x02\x18\x01\x1a_\n\nIndexValue\x12\x14\n\x08property\x18\x1e \x01(\tB\x02\x18\x01\x12;\n\x05value\x18\x1f \x02(\x0b\x32(.storage_onestore_v3_bytes.PropertyValueB\x02\x18\x01\":\n\x06\x43ursor\x12\x0e\n\x06\x63ursor\x18\x01 \x02(\x06\x12\x0b\n\x03\x61pp\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x61tabase_id\x18\x03 \x01(\t\"\x9f\x03\n\x05\x45rror\"\x95\x03\n\tErrorCode\x12\x0f\n\x0b\x42\x41\x44_REQUEST\x10\x01\x12\x1a\n\x16\x43ONCURRENT_TRANSACTION\x10\x02\x12\x12\n\x0eINTERNAL_ERROR\x10\x03\x12\x0e\n\nNEED_INDEX\x10\x04\x12\x0b\n\x07TIMEOUT\x10\x05\x12\x15\n\x11PERMISSION_DENIED\x10\x06\x12\x12\n\x0e\x42IGTABLE_ERROR\x10\x07\x12 \n\x1c\x43OMMITTED_BUT_STILL_APPLYING\x10\x08\x12\x17\n\x13\x43\x41PABILITY_DISABLED\x10\t\x12\x19\n\x15TRY_ALTERNATE_BACKEND\x10\n\x12\x15\n\x11SAFE_TIME_TOO_OLD\x10\x0b\x12\x16\n\x12RESOURCE_EXHAUSTED\x10\x0c\x12\x1c\n\x18SNAPSHOT_VERSION_TOO_OLD\x10\x12\x12\r\n\tNOT_FOUND\x10\r\x12\x12\n\x0e\x41LREADY_EXISTS\x10\x0e\x12\x17\n\x13\x46\x41ILED_PRECONDITION\x10\x0f\x12\x13\n\x0fUNAUTHENTICATED\x10\x10\x12\x0b\n\x07\x41\x42ORTED\x10\x11\"\xbd\x02\n\x04\x43ost\x12\x14\n\x0cindex_writes\x18\x01 \x01(\x05\x12\x19\n\x11index_write_bytes\x18\x02 \x01(\x05\x12\x15\n\rentity_writes\x18\x03 \x01(\x05\x12\x1a\n\x12\x65ntity_write_bytes\x18\x04 \x01(\x05\x12\x42\n\ncommitcost\x18\x05 \x01(\n2..apphosting_datastore_v3_bytes.Cost.CommitCost\x12!\n\x19\x61pproximate_storage_delta\x18\x08 \x01(\x05\x12\x1b\n\x13id_sequence_updates\x18\t \x01(\x05\x1aM\n\nCommitCost\x12\x1d\n\x15requested_entity_puts\x18\x06 \x01(\x05\x12 \n\x18requested_entity_deletes\x18\x07 \x01(\x05\"\xe4\x01\n\nGetRequest\x12\x31\n\x03key\x18\x01 \x03(\x0b\x32$.storage_onestore_v3_bytes.Reference\x12?\n\x0btransaction\x18\x02 \x01(\x0b\x32*.apphosting_datastore_v3_bytes.Transaction\x12\x13\n\x0b\x66\x61ilover_ms\x18\x03 \x01(\x03\x12\x0e\n\x06strong\x18\x04 \x01(\x08\x12\x1d\n\x0e\x61llow_deferred\x18\x05 \x01(\x08:\x05\x66\x61lse\x12\x1e\n\x16read_time_epoch_micros\x18\x06 \x01(\x03\"\xe9\x02\n\x0bGetResponse\x12\x41\n\x06\x65ntity\x18\x01 \x03(\n21.apphosting_datastore_v3_bytes.GetResponse.Entity\x12\x36\n\x08\x64\x65\x66\x65rred\x18\x05 \x03(\x0b\x32$.storage_onestore_v3_bytes.Reference\x12\x16\n\x08in_order\x18\x06 \x01(\x08:\x04true\x12\x1e\n\x16read_time_epoch_micros\x18\x07 \x01(\x03\x1a\xa6\x01\n\x06\x45ntity\x12\x36\n\x06\x65ntity\x18\x02 \x01(\x0b\x32&.storage_onestore_v3_bytes.EntityProto\x12\x31\n\x03key\x18\x04 \x01(\x0b\x32$.storage_onestore_v3_bytes.Reference\x12\x0f\n\x07version\x18\x03 \x01(\x03\x12 \n\x18update_time_epoch_micros\x18\x08 \x01(\x03\"\xe4\x03\n\nPutRequest\x12\x36\n\x06\x65ntity\x18\x01 \x03(\x0b\x32&.storage_onestore_v3_bytes.EntityProto\x12?\n\x0btransaction\x18\x02 \x01(\x0b\x32*.apphosting_datastore_v3_bytes.Transaction\x12\x42\n\x0f\x63omposite_index\x18\x03 \x03(\x0b\x32).storage_onestore_v3_bytes.CompositeIndex\x12\x16\n\x07trusted\x18\x04 \x01(\x08:\x05\x66\x61lse\x12\x14\n\x05\x66orce\x18\x07 \x01(\x08:\x05\x66\x61lse\x12\x1b\n\x0cmark_changes\x18\x08 \x01(\x08:\x05\x66\x61lse\x12/\n\x08snapshot\x18\t \x03(\x0b\x32\x1d.storage_onestore_v3.Snapshot\x12W\n\x0e\x61uto_id_policy\x18\n \x01(\x0e\x32\x36.apphosting_datastore_v3_bytes.PutRequest.AutoIdPolicy:\x07\x43URRENT\x12\x17\n\x0fsequence_number\x18\x0c \x01(\x03\"+\n\x0c\x41utoIdPolicy\x12\x0b\n\x07\x43URRENT\x10\x00\x12\x0e\n\nSEQUENTIAL\x10\x01\"\xa6\x01\n\x0bPutResponse\x12\x31\n\x03key\x18\x01 \x03(\x0b\x32$.storage_onestore_v3_bytes.Reference\x12\x31\n\x04\x63ost\x18\x02 \x01(\x0b\x32#.apphosting_datastore_v3_bytes.Cost\x12\x0f\n\x07version\x18\x03 \x03(\x03\x12 \n\x18update_time_epoch_micros\x18\x04 \x03(\x03\"\xcc\x01\n\x0cTouchRequest\x12\x31\n\x03key\x18\x01 \x03(\x0b\x32$.storage_onestore_v3_bytes.Reference\x12\x42\n\x0f\x63omposite_index\x18\x02 \x03(\x0b\x32).storage_onestore_v3_bytes.CompositeIndex\x12\x14\n\x05\x66orce\x18\x03 \x01(\x08:\x05\x66\x61lse\x12/\n\x08snapshot\x18\t \x03(\x0b\x32\x1d.storage_onestore_v3.Snapshot\"B\n\rTouchResponse\x12\x31\n\x04\x63ost\x18\x01 \x01(\x0b\x32#.apphosting_datastore_v3_bytes.Cost\"\xdc\x02\n\rDeleteRequest\x12\x31\n\x03key\x18\x06 \x03(\x0b\x32$.storage_onestore_v3_bytes.Reference\x12?\n\x0btransaction\x18\x05 \x01(\x0b\x32*.apphosting_datastore_v3_bytes.Transaction\x12\x42\n\x0f\x63omposite_index\x18\x0b \x03(\x0b\x32).storage_onestore_v3_bytes.CompositeIndex\x12\x16\n\x07trusted\x18\x04 \x01(\x08:\x05\x66\x61lse\x12\x14\n\x05\x66orce\x18\x07 \x01(\x08:\x05\x66\x61lse\x12\x1b\n\x0cmark_changes\x18\x08 \x01(\x08:\x05\x66\x61lse\x12/\n\x08snapshot\x18\t \x03(\x0b\x32\x1d.storage_onestore_v3.Snapshot\x12\x17\n\x0fsequence_number\x18\x0c \x01(\x03\"v\n\x0e\x44\x65leteResponse\x12\x31\n\x04\x63ost\x18\x01 \x01(\x0b\x32#.apphosting_datastore_v3_bytes.Cost\x12\x0f\n\x07version\x18\x03 \x03(\x03\x12 \n\x18\x64\x65lete_time_epoch_micros\x18\x04 \x03(\x03\"\x82\x01\n\x0bNextRequest\x12\x35\n\x06\x63ursor\x18\x01 \x02(\x0b\x32%.apphosting_datastore_v3_bytes.Cursor\x12\r\n\x05\x63ount\x18\x02 \x01(\x05\x12\x15\n\x06offset\x18\x04 \x01(\x05:\x01\x30\x42\x02\x18\x01\x12\x16\n\x07\x63ompile\x18\x03 \x01(\x08:\x05\x66\x61lse\"\xa7\x05\n\x0bQueryResult\x12\x35\n\x06\x63ursor\x18\x01 \x01(\x0b\x32%.apphosting_datastore_v3_bytes.Cursor\x12\x36\n\x06result\x18\x02 \x03(\x0b\x32&.storage_onestore_v3_bytes.EntityProto\x12\x17\n\x0fskipped_results\x18\x07 \x01(\x05\x12\x14\n\x0cmore_results\x18\x03 \x02(\x08\x12\x11\n\tkeys_only\x18\x04 \x01(\x08\x12\x12\n\nindex_only\x18\t \x01(\x08\x12\x11\n\tsmall_ops\x18\n \x01(\x08\x12\x44\n\x0e\x63ompiled_query\x18\x05 \x01(\x0b\x32,.apphosting_datastore_v3_bytes.CompiledQuery\x12\x46\n\x0f\x63ompiled_cursor\x18\x06 \x01(\x0b\x32-.apphosting_datastore_v3_bytes.CompiledCursor\x12\x38\n\x05index\x18\x08 \x03(\x0b\x32).storage_onestore_v3_bytes.CompositeIndex\x12\x0f\n\x07version\x18\x0b \x03(\x03\x12 \n\x18update_time_epoch_micros\x18\x0f \x03(\x03\x12\x1e\n\x16read_time_epoch_micros\x18\x0e \x01(\x03\x12M\n\x16result_compiled_cursor\x18\x0c \x03(\x0b\x32-.apphosting_datastore_v3_bytes.CompiledCursor\x12V\n\x1fskipped_results_compiled_cursor\x18\r \x01(\x0b\x32-.apphosting_datastore_v3_bytes.CompiledCursor\"\xb7\x01\n\x12\x41llocateIdsRequest\x12\x37\n\tmodel_key\x18\x01 \x01(\x0b\x32$.storage_onestore_v3_bytes.Reference\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x0b\n\x03max\x18\x03 \x01(\x03\x12\x35\n\x07reserve\x18\x05 \x03(\x0b\x32$.storage_onestore_v3_bytes.Reference\x12\x16\n\x07trusted\x18\x06 \x01(\x08:\x05\x66\x61lse\"d\n\x13\x41llocateIdsResponse\x12\r\n\x05start\x18\x01 \x02(\x03\x12\x0b\n\x03\x65nd\x18\x02 \x02(\x03\x12\x31\n\x04\x63ost\x18\x03 \x01(\x0b\x32#.apphosting_datastore_v3_bytes.Cost\"L\n\x10\x43ompositeIndices\x12\x38\n\x05index\x18\x01 \x03(\x0b\x32).storage_onestore_v3_bytes.CompositeIndex\"\x81\x01\n\x11\x41\x64\x64\x41\x63tionsRequest\x12?\n\x0btransaction\x18\x01 \x02(\x0b\x32*.apphosting_datastore_v3_bytes.Transaction\x12+\n\x06\x61\x63tion\x18\x02 \x03(\x0b\x32\x1b.storage_onestore_v3.Action\"\x14\n\x12\x41\x64\x64\x41\x63tionsResponse\"\xc5\x02\n\x17\x42\x65ginTransactionRequest\x12\x0b\n\x03\x61pp\x18\x01 \x02(\t\x12 \n\x11\x61llow_multiple_eg\x18\x02 \x01(\x08:\x05\x66\x61lse\x12\x13\n\x0b\x64\x61tabase_id\x18\x04 \x01(\t\x12]\n\x04mode\x18\x05 \x01(\x0e\x32\x46.apphosting_datastore_v3_bytes.BeginTransactionRequest.TransactionMode:\x07UNKNOWN\x12H\n\x14previous_transaction\x18\x07 \x01(\x0b\x32*.apphosting_datastore_v3_bytes.Transaction\"=\n\x0fTransactionMode\x12\x0b\n\x07UNKNOWN\x10\x00\x12\r\n\tREAD_ONLY\x10\x01\x12\x0e\n\nREAD_WRITE\x10\x02\"\xaa\x02\n\x0e\x43ommitResponse\x12\x31\n\x04\x63ost\x18\x01 \x01(\x0b\x32#.apphosting_datastore_v3_bytes.Cost\x12 \n\x18\x63ommit_time_epoch_micros\x18\x07 \x01(\x03\x12\x46\n\x07version\x18\x03 \x03(\n25.apphosting_datastore_v3_bytes.CommitResponse.Version\x1a{\n\x07Version\x12=\n\x0froot_entity_key\x18\x04 \x02(\x0b\x32$.storage_onestore_v3_bytes.Reference\x12\x0f\n\x07version\x18\x05 \x02(\x03\x12 \n\x18update_time_epoch_micros\x18\x06 \x01(\x03\"8\n\x11GetIndicesRequest\x12\x0e\n\x06\x61pp_id\x18\x01 \x02(\t\x12\x13\n\x0b\x64\x61tabase_id\x18\x02 \x01(\t\"6\n\x13UpdateIndexResponse\x12\x10\n\x08type_url\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c\"\x86\x02\n\x12\x44\x61tastoreService_3\"\xef\x01\n\x06Method\x12\x07\n\x03Get\x10\x01\x12\x07\n\x03Put\x10\x02\x12\t\n\x05Touch\x10\x03\x12\n\n\x06\x44\x65lete\x10\x04\x12\x0c\n\x08RunQuery\x10\x05\x12\x0e\n\nAddActions\x10\x06\x12\x08\n\x04Next\x10\x07\x12\x10\n\x0c\x44\x65leteCursor\x10\x08\x12\x14\n\x10\x42\x65ginTransaction\x10\t\x12\n\n\x06\x43ommit\x10\n\x12\x0c\n\x08Rollback\x10\x0b\x12\x0f\n\x0b\x41llocateIds\x10\x0c\x12\x0f\n\x0b\x43reateIndex\x10\r\x12\x0f\n\x0bUpdateIndex\x10\x0e\x12\x0e\n\nGetIndices\x10\x0f\x12\x0f\n\x0b\x44\x65leteIndex\x10\x10\x42\x36\n%com.google.google.appengine.datastoreB\rDatastoreV3Pb')



_TRANSACTION = DESCRIPTOR.message_types_by_name['Transaction']
_QUERY = DESCRIPTOR.message_types_by_name['Query']
_QUERY_FILTER = _QUERY.nested_types_by_name['Filter']
_QUERY_ORDER = _QUERY.nested_types_by_name['Order']
_REGIONPOINT = DESCRIPTOR.message_types_by_name['RegionPoint']
_CIRCLEREGION = DESCRIPTOR.message_types_by_name['CircleRegion']
_RECTANGLEREGION = DESCRIPTOR.message_types_by_name['RectangleRegion']
_GEOREGION = DESCRIPTOR.message_types_by_name['GeoRegion']
_COMPILEDQUERY = DESCRIPTOR.message_types_by_name['CompiledQuery']
_COMPILEDQUERY_PRIMARYSCAN = _COMPILEDQUERY.nested_types_by_name['PrimaryScan']
_COMPILEDQUERY_MERGEJOINSCAN = _COMPILEDQUERY.nested_types_by_name['MergeJoinScan']
_COMPILEDQUERY_ENTITYFILTER = _COMPILEDQUERY.nested_types_by_name['EntityFilter']
_COMPILEDCURSOR = DESCRIPTOR.message_types_by_name['CompiledCursor']
_COMPILEDCURSOR_POSITION = _COMPILEDCURSOR.nested_types_by_name['Position']
_COMPILEDCURSOR_POSITION_INDEXVALUE = _COMPILEDCURSOR_POSITION.nested_types_by_name['IndexValue']
_CURSOR = DESCRIPTOR.message_types_by_name['Cursor']
_ERROR = DESCRIPTOR.message_types_by_name['Error']
_COST = DESCRIPTOR.message_types_by_name['Cost']
_COST_COMMITCOST = _COST.nested_types_by_name['CommitCost']
_GETREQUEST = DESCRIPTOR.message_types_by_name['GetRequest']
_GETRESPONSE = DESCRIPTOR.message_types_by_name['GetResponse']
_GETRESPONSE_ENTITY = _GETRESPONSE.nested_types_by_name['Entity']
_PUTREQUEST = DESCRIPTOR.message_types_by_name['PutRequest']
_PUTRESPONSE = DESCRIPTOR.message_types_by_name['PutResponse']
_TOUCHREQUEST = DESCRIPTOR.message_types_by_name['TouchRequest']
_TOUCHRESPONSE = DESCRIPTOR.message_types_by_name['TouchResponse']
_DELETEREQUEST = DESCRIPTOR.message_types_by_name['DeleteRequest']
_DELETERESPONSE = DESCRIPTOR.message_types_by_name['DeleteResponse']
_NEXTREQUEST = DESCRIPTOR.message_types_by_name['NextRequest']
_QUERYRESULT = DESCRIPTOR.message_types_by_name['QueryResult']
_ALLOCATEIDSREQUEST = DESCRIPTOR.message_types_by_name['AllocateIdsRequest']
_ALLOCATEIDSRESPONSE = DESCRIPTOR.message_types_by_name['AllocateIdsResponse']
_COMPOSITEINDICES = DESCRIPTOR.message_types_by_name['CompositeIndices']
_ADDACTIONSREQUEST = DESCRIPTOR.message_types_by_name['AddActionsRequest']
_ADDACTIONSRESPONSE = DESCRIPTOR.message_types_by_name['AddActionsResponse']
_BEGINTRANSACTIONREQUEST = DESCRIPTOR.message_types_by_name['BeginTransactionRequest']
_COMMITRESPONSE = DESCRIPTOR.message_types_by_name['CommitResponse']
_COMMITRESPONSE_VERSION = _COMMITRESPONSE.nested_types_by_name['Version']
_GETINDICESREQUEST = DESCRIPTOR.message_types_by_name['GetIndicesRequest']
_UPDATEINDEXRESPONSE = DESCRIPTOR.message_types_by_name['UpdateIndexResponse']
_DATASTORESERVICE_3 = DESCRIPTOR.message_types_by_name['DatastoreService_3']
_QUERY_FILTER_OPERATOR = _QUERY_FILTER.enum_types_by_name['Operator']
_QUERY_ORDER_DIRECTION = _QUERY_ORDER.enum_types_by_name['Direction']
_QUERY_HINT = _QUERY.enum_types_by_name['Hint']
_ERROR_ERRORCODE = _ERROR.enum_types_by_name['ErrorCode']
_PUTREQUEST_AUTOIDPOLICY = _PUTREQUEST.enum_types_by_name['AutoIdPolicy']
_BEGINTRANSACTIONREQUEST_TRANSACTIONMODE = _BEGINTRANSACTIONREQUEST.enum_types_by_name['TransactionMode']
_DATASTORESERVICE_3_METHOD = _DATASTORESERVICE_3.enum_types_by_name['Method']
Transaction = _reflection.GeneratedProtocolMessageType('Transaction', (_message.Message,), {
  'DESCRIPTOR' : _TRANSACTION,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(Transaction)

Query = _reflection.GeneratedProtocolMessageType('Query', (_message.Message,), {

  'Filter' : _reflection.GeneratedProtocolMessageType('Filter', (_message.Message,), {
    'DESCRIPTOR' : _QUERY_FILTER,
    '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

    })
  ,

  'Order' : _reflection.GeneratedProtocolMessageType('Order', (_message.Message,), {
    'DESCRIPTOR' : _QUERY_ORDER,
    '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

    })
  ,
  'DESCRIPTOR' : _QUERY,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(Query)
_sym_db.RegisterMessage(Query.Filter)
_sym_db.RegisterMessage(Query.Order)

RegionPoint = _reflection.GeneratedProtocolMessageType('RegionPoint', (_message.Message,), {
  'DESCRIPTOR' : _REGIONPOINT,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(RegionPoint)

CircleRegion = _reflection.GeneratedProtocolMessageType('CircleRegion', (_message.Message,), {
  'DESCRIPTOR' : _CIRCLEREGION,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(CircleRegion)

RectangleRegion = _reflection.GeneratedProtocolMessageType('RectangleRegion', (_message.Message,), {
  'DESCRIPTOR' : _RECTANGLEREGION,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(RectangleRegion)

GeoRegion = _reflection.GeneratedProtocolMessageType('GeoRegion', (_message.Message,), {
  'DESCRIPTOR' : _GEOREGION,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(GeoRegion)

CompiledQuery = _reflection.GeneratedProtocolMessageType('CompiledQuery', (_message.Message,), {

  'PrimaryScan' : _reflection.GeneratedProtocolMessageType('PrimaryScan', (_message.Message,), {
    'DESCRIPTOR' : _COMPILEDQUERY_PRIMARYSCAN,
    '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

    })
  ,

  'MergeJoinScan' : _reflection.GeneratedProtocolMessageType('MergeJoinScan', (_message.Message,), {
    'DESCRIPTOR' : _COMPILEDQUERY_MERGEJOINSCAN,
    '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

    })
  ,

  'EntityFilter' : _reflection.GeneratedProtocolMessageType('EntityFilter', (_message.Message,), {
    'DESCRIPTOR' : _COMPILEDQUERY_ENTITYFILTER,
    '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

    })
  ,
  'DESCRIPTOR' : _COMPILEDQUERY,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(CompiledQuery)
_sym_db.RegisterMessage(CompiledQuery.PrimaryScan)
_sym_db.RegisterMessage(CompiledQuery.MergeJoinScan)
_sym_db.RegisterMessage(CompiledQuery.EntityFilter)

CompiledCursor = _reflection.GeneratedProtocolMessageType('CompiledCursor', (_message.Message,), {

  'Position' : _reflection.GeneratedProtocolMessageType('Position', (_message.Message,), {

    'IndexValue' : _reflection.GeneratedProtocolMessageType('IndexValue', (_message.Message,), {
      'DESCRIPTOR' : _COMPILEDCURSOR_POSITION_INDEXVALUE,
      '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

      })
    ,
    'DESCRIPTOR' : _COMPILEDCURSOR_POSITION,
    '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

    })
  ,
  'DESCRIPTOR' : _COMPILEDCURSOR,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(CompiledCursor)
_sym_db.RegisterMessage(CompiledCursor.Position)
_sym_db.RegisterMessage(CompiledCursor.Position.IndexValue)

Cursor = _reflection.GeneratedProtocolMessageType('Cursor', (_message.Message,), {
  'DESCRIPTOR' : _CURSOR,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(Cursor)

Error = _reflection.GeneratedProtocolMessageType('Error', (_message.Message,), {
  'DESCRIPTOR' : _ERROR,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(Error)

Cost = _reflection.GeneratedProtocolMessageType('Cost', (_message.Message,), {

  'CommitCost' : _reflection.GeneratedProtocolMessageType('CommitCost', (_message.Message,), {
    'DESCRIPTOR' : _COST_COMMITCOST,
    '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

    })
  ,
  'DESCRIPTOR' : _COST,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(Cost)
_sym_db.RegisterMessage(Cost.CommitCost)

GetRequest = _reflection.GeneratedProtocolMessageType('GetRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETREQUEST,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(GetRequest)

GetResponse = _reflection.GeneratedProtocolMessageType('GetResponse', (_message.Message,), {

  'Entity' : _reflection.GeneratedProtocolMessageType('Entity', (_message.Message,), {
    'DESCRIPTOR' : _GETRESPONSE_ENTITY,
    '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

    })
  ,
  'DESCRIPTOR' : _GETRESPONSE,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(GetResponse)
_sym_db.RegisterMessage(GetResponse.Entity)

PutRequest = _reflection.GeneratedProtocolMessageType('PutRequest', (_message.Message,), {
  'DESCRIPTOR' : _PUTREQUEST,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(PutRequest)

PutResponse = _reflection.GeneratedProtocolMessageType('PutResponse', (_message.Message,), {
  'DESCRIPTOR' : _PUTRESPONSE,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(PutResponse)

TouchRequest = _reflection.GeneratedProtocolMessageType('TouchRequest', (_message.Message,), {
  'DESCRIPTOR' : _TOUCHREQUEST,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(TouchRequest)

TouchResponse = _reflection.GeneratedProtocolMessageType('TouchResponse', (_message.Message,), {
  'DESCRIPTOR' : _TOUCHRESPONSE,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(TouchResponse)

DeleteRequest = _reflection.GeneratedProtocolMessageType('DeleteRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEREQUEST,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(DeleteRequest)

DeleteResponse = _reflection.GeneratedProtocolMessageType('DeleteResponse', (_message.Message,), {
  'DESCRIPTOR' : _DELETERESPONSE,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(DeleteResponse)

NextRequest = _reflection.GeneratedProtocolMessageType('NextRequest', (_message.Message,), {
  'DESCRIPTOR' : _NEXTREQUEST,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(NextRequest)

QueryResult = _reflection.GeneratedProtocolMessageType('QueryResult', (_message.Message,), {
  'DESCRIPTOR' : _QUERYRESULT,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(QueryResult)

AllocateIdsRequest = _reflection.GeneratedProtocolMessageType('AllocateIdsRequest', (_message.Message,), {
  'DESCRIPTOR' : _ALLOCATEIDSREQUEST,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(AllocateIdsRequest)

AllocateIdsResponse = _reflection.GeneratedProtocolMessageType('AllocateIdsResponse', (_message.Message,), {
  'DESCRIPTOR' : _ALLOCATEIDSRESPONSE,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(AllocateIdsResponse)

CompositeIndices = _reflection.GeneratedProtocolMessageType('CompositeIndices', (_message.Message,), {
  'DESCRIPTOR' : _COMPOSITEINDICES,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(CompositeIndices)

AddActionsRequest = _reflection.GeneratedProtocolMessageType('AddActionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _ADDACTIONSREQUEST,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(AddActionsRequest)

AddActionsResponse = _reflection.GeneratedProtocolMessageType('AddActionsResponse', (_message.Message,), {
  'DESCRIPTOR' : _ADDACTIONSRESPONSE,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(AddActionsResponse)

BeginTransactionRequest = _reflection.GeneratedProtocolMessageType('BeginTransactionRequest', (_message.Message,), {
  'DESCRIPTOR' : _BEGINTRANSACTIONREQUEST,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(BeginTransactionRequest)

CommitResponse = _reflection.GeneratedProtocolMessageType('CommitResponse', (_message.Message,), {

  'Version' : _reflection.GeneratedProtocolMessageType('Version', (_message.Message,), {
    'DESCRIPTOR' : _COMMITRESPONSE_VERSION,
    '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

    })
  ,
  'DESCRIPTOR' : _COMMITRESPONSE,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(CommitResponse)
_sym_db.RegisterMessage(CommitResponse.Version)

GetIndicesRequest = _reflection.GeneratedProtocolMessageType('GetIndicesRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETINDICESREQUEST,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(GetIndicesRequest)

UpdateIndexResponse = _reflection.GeneratedProtocolMessageType('UpdateIndexResponse', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEINDEXRESPONSE,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(UpdateIndexResponse)

DatastoreService_3 = _reflection.GeneratedProtocolMessageType('DatastoreService_3', (_message.Message,), {
  'DESCRIPTOR' : _DATASTORESERVICE_3,
  '__module__' : 'google.appengine.datastore.datastore_v3_bytes_pb2'

  })
_sym_db.RegisterMessage(DatastoreService_3)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n%com.google.google.appengine.datastoreB\rDatastoreV3Pb'
  _QUERY.fields_by_name['read_time_us']._options = None
  _QUERY.fields_by_name['read_time_us']._serialized_options = b'\030\001'
  _QUERY.fields_by_name['hint']._options = None
  _QUERY.fields_by_name['hint']._serialized_options = b'\030\001'
  _QUERY.fields_by_name['search_query']._options = None
  _QUERY.fields_by_name['search_query']._serialized_options = b'\030\001'
  _QUERY.fields_by_name['require_perfect_plan']._options = None
  _QUERY.fields_by_name['require_perfect_plan']._serialized_options = b'\030\001'
  _QUERY.fields_by_name['compile']._options = None
  _QUERY.fields_by_name['compile']._serialized_options = b'\030\001'
  _QUERY.fields_by_name['persist_offset']._options = None
  _QUERY.fields_by_name['persist_offset']._serialized_options = b'\030\001'
  _COMPILEDCURSOR_POSITION_INDEXVALUE.fields_by_name['property']._options = None
  _COMPILEDCURSOR_POSITION_INDEXVALUE.fields_by_name['property']._serialized_options = b'\030\001'
  _COMPILEDCURSOR_POSITION_INDEXVALUE.fields_by_name['value']._options = None
  _COMPILEDCURSOR_POSITION_INDEXVALUE.fields_by_name['value']._serialized_options = b'\030\001'
  _COMPILEDCURSOR_POSITION.fields_by_name['start_key']._options = None
  _COMPILEDCURSOR_POSITION.fields_by_name['start_key']._serialized_options = b'\030\001'
  _COMPILEDCURSOR_POSITION.fields_by_name['indexvalue']._options = None
  _COMPILEDCURSOR_POSITION.fields_by_name['indexvalue']._serialized_options = b'\030\001'
  _COMPILEDCURSOR_POSITION.fields_by_name['key']._options = None
  _COMPILEDCURSOR_POSITION.fields_by_name['key']._serialized_options = b'\030\001'
  _COMPILEDCURSOR_POSITION.fields_by_name['start_inclusive']._options = None
  _COMPILEDCURSOR_POSITION.fields_by_name['start_inclusive']._serialized_options = b'\030\001'
  _COMPILEDCURSOR_POSITION.fields_by_name['before_ascending']._options = None
  _COMPILEDCURSOR_POSITION.fields_by_name['before_ascending']._serialized_options = b'\030\001'
  _COMPILEDCURSOR.fields_by_name['position']._options = None
  _COMPILEDCURSOR.fields_by_name['position']._serialized_options = b'\030\001'
  _NEXTREQUEST.fields_by_name['offset']._options = None
  _NEXTREQUEST.fields_by_name['offset']._serialized_options = b'\030\001'
  _TRANSACTION._serialized_start=218
  _TRANSACTION._serialized_end=378
  _QUERY._serialized_start=381
  _QUERY._serialized_end=2014
  _QUERY_FILTER._serialized_start=1424
  _QUERY_FILTER._serialized_end=1795
  _QUERY_FILTER_OPERATOR._serialized_start=1618
  _QUERY_FILTER_OPERATOR._serialized_end=1795
  _QUERY_ORDER._serialized_start=1798
  _QUERY_ORDER._serialized_end=1951
  _QUERY_ORDER_DIRECTION._serialized_start=1909
  _QUERY_ORDER_DIRECTION._serialized_end=1951
  _QUERY_HINT._serialized_start=1953
  _QUERY_HINT._serialized_end=2014
  _REGIONPOINT._serialized_start=2016
  _REGIONPOINT._serialized_end=2066
  _CIRCLEREGION._serialized_start=2068
  _CIRCLEREGION._serialized_end=2165
  _RECTANGLEREGION._serialized_start=2168
  _RECTANGLEREGION._serialized_end=2311
  _GEOREGION._serialized_start=2314
  _GEOREGION._serialized_end=2453
  _COMPILEDQUERY._serialized_start=2456
  _COMPILEDQUERY._serialized_end=3332
  _COMPILEDQUERY_PRIMARYSCAN._serialized_start=2920
  _COMPILEDQUERY_PRIMARYSCAN._serialized_end=3133
  _COMPILEDQUERY_MERGEJOINSCAN._serialized_start=3135
  _COMPILEDQUERY_MERGEJOINSCAN._serialized_end=3221
  _COMPILEDQUERY_ENTITYFILTER._serialized_start=3223
  _COMPILEDQUERY_ENTITYFILTER._serialized_end=3332
  _COMPILEDCURSOR._serialized_start=3335
  _COMPILEDCURSOR._serialized_end=3909
  _COMPILEDCURSOR_POSITION._serialized_start=3568
  _COMPILEDCURSOR_POSITION._serialized_end=3909
  _COMPILEDCURSOR_POSITION_INDEXVALUE._serialized_start=3814
  _COMPILEDCURSOR_POSITION_INDEXVALUE._serialized_end=3909
  _CURSOR._serialized_start=3911
  _CURSOR._serialized_end=3969
  _ERROR._serialized_start=3972
  _ERROR._serialized_end=4387
  _ERROR_ERRORCODE._serialized_start=3982
  _ERROR_ERRORCODE._serialized_end=4387
  _COST._serialized_start=4390
  _COST._serialized_end=4707
  _COST_COMMITCOST._serialized_start=4630
  _COST_COMMITCOST._serialized_end=4707
  _GETREQUEST._serialized_start=4710
  _GETREQUEST._serialized_end=4938
  _GETRESPONSE._serialized_start=4941
  _GETRESPONSE._serialized_end=5302
  _GETRESPONSE_ENTITY._serialized_start=5136
  _GETRESPONSE_ENTITY._serialized_end=5302
  _PUTREQUEST._serialized_start=5305
  _PUTREQUEST._serialized_end=5789
  _PUTREQUEST_AUTOIDPOLICY._serialized_start=5746
  _PUTREQUEST_AUTOIDPOLICY._serialized_end=5789
  _PUTRESPONSE._serialized_start=5792
  _PUTRESPONSE._serialized_end=5958
  _TOUCHREQUEST._serialized_start=5961
  _TOUCHREQUEST._serialized_end=6165
  _TOUCHRESPONSE._serialized_start=6167
  _TOUCHRESPONSE._serialized_end=6233
  _DELETEREQUEST._serialized_start=6236
  _DELETEREQUEST._serialized_end=6584
  _DELETERESPONSE._serialized_start=6586
  _DELETERESPONSE._serialized_end=6704
  _NEXTREQUEST._serialized_start=6707
  _NEXTREQUEST._serialized_end=6837
  _QUERYRESULT._serialized_start=6840
  _QUERYRESULT._serialized_end=7519
  _ALLOCATEIDSREQUEST._serialized_start=7522
  _ALLOCATEIDSREQUEST._serialized_end=7705
  _ALLOCATEIDSRESPONSE._serialized_start=7707
  _ALLOCATEIDSRESPONSE._serialized_end=7807
  _COMPOSITEINDICES._serialized_start=7809
  _COMPOSITEINDICES._serialized_end=7885
  _ADDACTIONSREQUEST._serialized_start=7888
  _ADDACTIONSREQUEST._serialized_end=8017
  _ADDACTIONSRESPONSE._serialized_start=8019
  _ADDACTIONSRESPONSE._serialized_end=8039
  _BEGINTRANSACTIONREQUEST._serialized_start=8042
  _BEGINTRANSACTIONREQUEST._serialized_end=8367
  _BEGINTRANSACTIONREQUEST_TRANSACTIONMODE._serialized_start=8306
  _BEGINTRANSACTIONREQUEST_TRANSACTIONMODE._serialized_end=8367
  _COMMITRESPONSE._serialized_start=8370
  _COMMITRESPONSE._serialized_end=8668
  _COMMITRESPONSE_VERSION._serialized_start=8545
  _COMMITRESPONSE_VERSION._serialized_end=8668
  _GETINDICESREQUEST._serialized_start=8670
  _GETINDICESREQUEST._serialized_end=8726
  _UPDATEINDEXRESPONSE._serialized_start=8728
  _UPDATEINDEXRESPONSE._serialized_end=8782
  _DATASTORESERVICE_3._serialized_start=8785
  _DATASTORESERVICE_3._serialized_end=9047
  _DATASTORESERVICE_3_METHOD._serialized_start=8808
  _DATASTORESERVICE_3_METHOD._serialized_end=9047




#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""A request-local environment and logging stream."""








import abc
import cgi
import collections
import os
import sys
import threading

import six

_sys_stderr = sys.stderr


class RequestEnvironment(threading.local):
  """A thread local request environment.

  A thread local request environment that provides an error stream errors and a
  dict of the request environment environ as would be found in os.environ. A
  single error stream is shared between threads of a single request, but each
  thread accesses an independent copy of environ created when
  CloneRequestEnvironment is called. A request environment of one thread can be
  installed in another thread as follows:
    1. Call CloneRequestEnvironment in the first thread.
    2. Call the returned callable from the other thread.
  """

  def __init__(self):
    super(RequestEnvironment, self).__init__()
    self.Reset()

  def Reset(self):
    """Resets the error stream and environment for this request."""
    self.errors = _sys_stderr
    self.environ = {}
    self.runtime_stats = {}

  def Init(self, errors, environ, runtime_stats=None):
    self.errors = errors

    if environ is not None:
      self.environ = environ

    if runtime_stats is None:
      self.runtime_stats = {}
    else:
      self.runtime_stats = runtime_stats

  def CloneRequestEnvironment(self):
    """Returns a callable that will install the environment in another thread.

    Returns:
      A callable that will duplicate the request environment of this thread in
      another thread that calls it.
    """
    errors = self.errors
    environ = dict(self.environ)
    runtime_stats = self.runtime_stats
    return lambda: self.Init(errors, environ, runtime_stats)

  def Clear(self):
    """Clears the thread locals."""
    self.__dict__.clear()
    self.Reset()

  def IncrementStat(self, metric, code):
    """Increments a tracked runtime statistic."""
    m = self.runtime_stats.get(metric)
    if m is None: return
    m[code] = m.get(code, 0) + 1


class RequestLocalStream(object):
  """A stream that delegates to a RequestEnvironment stream."""

  def __init__(self, request):
    self._request = request

  def close(self):
    pass

  def flush(self):
    self._request.errors.flush()

  def write(self, data):
    self._request.errors.write(data)

  def writelines(self, data):
    self._request.errors.writelines(data)

  def isatty(self):
    return False


class RequestLocalEnviron(collections.abc.MutableMapping):
  """A MutableMapping that delegates to a RequestEnvironment environ."""

  def __init__(self, request):
    self._request = request

  def __len__(self):
    return len(self._request.environ)

  def __iter__(self):
    return iter(self._request.environ)

  def __getitem__(self, key):
    return self._request.environ[key]

  def __setitem__(self, key, value):
    self._request.environ[key] = value

  def __delitem__(self, key):
    del self._request.environ[key]

  def __repr__(self):
    return repr(self._request.environ)

  def has_key(self, key):
    return key in self._request.environ

  def copy(self):
    return dict(self._request.environ)



  def viewitems(self):
    return collections.ItemsView(self)

  def viewkeys(self):
    return collections.KeysView(self)

  def viewvalues(self):
    return collections.ValuesView(self)


current_request = RequestEnvironment()





def PatchOsEnviron(os_module=os):
  """Replace os.environ by a RequestLocalEnviron instance.

  This is called from init.py when it modifies the execution
  environment (in the wider sense of the word).

  Args:
    os_module: An optional module to patch. Defaults to os.
  """
  os_module.environ = RequestLocalEnviron(current_request)


  cached_class = cgi.FieldStorage
  six.moves.reload_module(cgi)




  class FieldStorage(six.with_metaclass(abc.ABCMeta, cgi.FieldStorage, object)):
    pass

  FieldStorage.register(cached_class)

  cgi.FieldStorage = FieldStorage
